#!/usr/bin/env python3

import random
import sys
import math
from functools import reduce

# Code that shows how to attack a simple LCG PRNG by obtaining a
# series of direct outputs from the PRNG.
#
# Discussed in the following lecture:
# - https://teaching.leonardotamiano.xyz/cns/labs/lab_01_breaking_prng_part_1/
#
#
# Linear Congruential Generator
class LCG(object):
    def __init__(self, seed=1337, a=1103515245,b=12345,c=2147483648):   # generate seed with 56 bits of entropy
        self.x = seed
        self.a = a
        self.b = b
        self.c = c

    def next(self):
        self.x = (self.x * self.a + self.b) % self.c
        return self.x

    def int_32(self, number):
        return int(0xFFFFFFFF & number)    

# -------------------------------------

# Implements two different strategies for computing the modulus
# starting from a sequence of outputs. These two extra method use
# determinants and are based by research done by George Marsaglia.
#
# Useful links:
#
# - https://www.pnas.org/doi/10.1073/pnas.61.1.25
# - https://security.stackexchange.com/questions/4268/cracking-a-linear-congruential-generator
# 
def compute_det1(outputs, j):
    a = outputs[j] - outputs[0]
    b = outputs[j+1] - outputs[1]
    c = outputs[j+1] - outputs[0]
    d = outputs[j+2] - outputs[1]
    return abs(a*d - b*c)

def compute_modulus_extra_1(outputs):
    n = len(outputs)
    determinants = []
    for j in range(1, n-2):
        det = compute_det1(outputs, j)
        determinants.append(det)
    modulus =  reduce(math.gcd, determinants) #!
    return modulus

# -----

def compute_det2(val1, val2, val3, val4):
    return abs(val1*val3 - val1*val4 + val2*val3 - val2*val2 + val2*val4 - val3*val3)

def compute_modulus_extra_2(outputs):
    determinants = []
    
    n = len(outputs)
    for j in range(0, n-3):
        det = compute_det2(outputs[j], outputs[j+1], outputs[j+2], outputs[j+3])
        determinants.append(det)
    modulus =  reduce(math.gcd, determinants) #!
    return modulus

# -----

# https://security.stackexchange.com/questions/4268/cracking-a-linear-congruential-generator
def compute_modulus(outputs):
    ts = []
    for i in range(0, len(outputs) - 1):
        ts.append(outputs[i+1] - outputs[i])
        
    us = []
    for i in range(0, len(ts)-2):
        us.append(abs(ts[i+2]*ts[i] - ts[i+1]**2))

    modulus =  reduce(math.gcd, us) #!
    return modulus

def compute_multiplier(outputs, modulus):
    term_1 = outputs[1] - outputs[2]
    term_2 = pow(outputs[0] - outputs[1], -1, modulus) #!
    a = (term_1 * term_2) % modulus
    return a

def compute_increment(outputs, modulus, a):
    b = (outputs[1] - outputs[0] * a) % modulus
    return b

# -------------------------------------    

def main():
    prng = LCG(seed=1337, a=1337, b=42, c=444447)

    n = 10
    outputs = []
    for i in range(0, n):
        outputs.append(prng.next())

    print(outputs)

    c = compute_modulus(outputs)
    a = compute_multiplier(outputs, c)
    b = compute_increment(outputs, c, a)

    print(f"c={c}")    
    print(f"a={a}")
    print(f"b={b}")
    
if __name__ == "__main__":
    main()
